/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007-2008 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: ConnectionManager.java 2977 2008-04-27 15:41:42Z benoitf $
 * --------------------------------------------------------------------------
 */
package org.ow2.util.cluster.jgroups;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.ChannelListener;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class manages connections on a JGroups object (DistributedTree,
 * Channel). Calls on certain methods of those JGroups objects are routed here.
 * If the object is not connected, reconnections are performed. There is a
 * reconnection time out.
 * @author eyindanga
 */
public class ConnectionManager implements ChannelListener, InvocationHandler {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ConnectionManager.class);

    /**
     * Reconnection timeout.
     */
    private volatile long timeout;

    /**
     * Manage the connection of this object.
     */
    private final Object handled;

    /**
     * Handle methods define in this interface.
     */
    private final Class<?> itfHandled;

    /**
     * true if handled object is connected.
     */
    private volatile boolean connected = true;

    public ConnectionManager(final long timeout, final Object handled, final Class<?> itfHandled) {
        super();
        this.timeout = timeout;
        this.handled = handled;
        this.itfHandled = itfHandled;
    }

    public ConnectionManager(final long timeout, final Object handled) {
        super();
        this.timeout = timeout;
        this.handled = handled;
        itfHandled = null;
    }

    public void channelClosed(final Channel arg0) {
        connected = false;

    }

    public void channelConnected(final Channel arg0) {
        connected = true;
        synchronized (this) {
            notifyAll();
        }
    }

    public void channelDisconnected(final Channel arg0) {
        connected = false;

    }

    public void channelReconnected(final Address arg0) {
        connected = true;
        synchronized (this) {
            notifyAll();
        }

    }

    public void channelShunned() {

    }

    public Object getJgroupsObject() {
        return handled;
    }

    public long getTimeout() {
        return timeout;
    }

    public void setTimeout(final long timeout) {
        this.timeout = timeout;
    }

    public boolean isConnected() {
        return connected;
    }

    public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {
        Object ret = null;

        if (handleThisMethod(method)) {
            long startTime = System.currentTimeMillis();
            while (!connected && System.currentTimeMillis() - startTime < timeout) {
                synchronized (this) {
                    wait(timeout);
                }
            }
            if (connected) {
                // invoke
                return method.invoke(handled, args);
            } else {
                throw new Exception(+timeout + " ms reconnection timeout exceeded, unable to invoke " + method.getName()
                        + " with args " + args + "\n");
            }

        }
        try {
            ret = method.invoke(handled, args);
        } catch (IllegalAccessException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        } catch (IllegalArgumentException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        } catch (InvocationTargetException e) {
            logger.error("Cannot invoke the method {0}", method.getName(), e);
        }
        return ret;
    }

    private boolean handleThisMethod(final Method method) {
        if (itfHandled == null) {
            // Handle all the methods excepted these defined in the class Object
            if (method.getDeclaringClass() == Object.class) {
                return false;
            } else {
                return true;
            }
        }
        Method[] handledMethods = itfHandled.getDeclaredMethods();
        for (int i = 0; i < handledMethods.length; i++) {
            // FIXME, compare number of parameters.
            if (method.getName().equals(handledMethods[i].getName())
                    && method.getParameterTypes().length == handledMethods[i].getParameterTypes().length) {
                return true;
            }
        }
        return false;
    }

}
